/*
 * Projet de lancer de rayons
 *
 * Jean-Guillaume Delculry
 *
 * MAIN.CPP
 *
 */

#include <iostream.h>
#include <stdio.h>
#include <math.h>
#include "structures.h"
#include<string.h>

#define PI	3.1415926
#define MAT 0
#define MIROIR 1


int nbsources;
Point *source;

int nbplans;
Plan *plan;

int nbspheres;
Sphere *sphere;

int nbfacettes;
Facette *facette;


ObjetIntersection CalculIntersections(Rayon,Plan *,Sphere *,Facette *);
ObjetIntersection CalculIntersectionsOmbres(Rayon,Sphere *,Facette *);
Luminosite raytrace(Rayon,int);


int main(int argc, char*argv[])
{
	int NP,NL;
	double ANGLE;
	double xoeil,yoeil,zoeil;
	double xdest,ydest,zdest;

	int res;
	double F;
	Luminosite Lum;
	Point Ooeil,Dest;
	Point Osphere_1,Osphere_2;
	Point PtEcranLocal,PtEcranGlobal,Pinter;
	Matrice MatPassage;
	Vecteur Vtmp;
	Rayon Ray;
	Image ima;
	int nbreflexionsuccesives;
	Point Rouge,Vert,Bleu,Violet,Abricot,Rose,Jaune,Noir,Blanc;
	char argimage[]="scene.ppm";
	char argscene[]="scene.txt";
	FILE *scenefile;
	char chaine[30];


	if ((scenefile = fopen(argscene, "r")) == NULL)   
	{   
		printf("Erreur d'ouverture du fichier texte en lecture\n");   
		return(0);   
	}   

	/*************************************************************
	 * Lecture dans le fichier de scene &
	 * Configuration du raytraceur
	 *************************************************************/
	fscanf(scenefile, "%s %d", &chaine, &NL);
	fscanf(scenefile, "%s %d", &chaine, &NP);
	fscanf(scenefile, "%s %lf", &chaine, &ANGLE);

	fscanf(scenefile, "%s %lf %lf %lf", &chaine, &xoeil, &yoeil, &zoeil);
	fscanf(scenefile, "%s %lf %lf %lf", &chaine, &xdest, &ydest, &zdest);

	/* SOURCES */
	fscanf(scenefile, "%s %d", &chaine, &nbsources);
	source = new Point[nbsources];
	for( int count=0; count<nbsources; count++ )
	{
		double x,y,z;
		fscanf(scenefile, "%lf %lf %lf", &x, &y, &z);
		source[count] = Point( x,y,z );
	}


	/* PLANS */
	fscanf(scenefile, "%s %d", &chaine, &nbplans);
	plan = new Plan[nbplans];
	for( count=0; count<nbplans; count++ )
	{
		double x,y,z,d;
		Point Coul;
		int miroir;
		fscanf(scenefile, "%lf %lf %lf %lf %s %lf %lf %lf %s %d",  &x, &y, &z, &d, &chaine,
																	&Coul.x, &Coul.y, &Coul.z, &chaine,
																	&miroir);
		plan[count] = Plan( Vecteur(x,y,z) , d, Coul, miroir);
	}


	/* SPHERES */
	fscanf(scenefile, "%s %d", &chaine, &nbspheres);
	sphere = new Sphere[nbspheres];
	for( count=0; count<nbspheres; count++ )
	{
		double x,y,z,r;
		Point Coul;
		int miroir;
		fscanf(scenefile, "%lf %lf %lf %lf %s %lf %lf %lf %s %d",  &x, &y, &z, &r, &chaine,
																	&Coul.x, &Coul.y, &Coul.z, &chaine,
																	&miroir);
		sphere[count] = Sphere( Point(x,y,z), r, Coul, miroir);
	}


	/* FACETTES */
	fscanf(scenefile, "%s %d", &chaine, &nbfacettes);
	facette = new Facette[nbfacettes];
	for( count=0; count<nbfacettes; count++ )
	{

		double x1,y1,z1,x2,y2,z2,x3,y3,z3;
		Point Coul;
		int miroir;
		fscanf(scenefile, "%lf %lf %lf %lf %lf %lf %lf %lf %lf %s %lf %lf %lf %s %d",
																	&x1, &y1, &z1,
																	&x2, &y2, &z2,
																	&x3, &y3, &z3, &chaine,
																	&Coul.x, &Coul.y, &Coul.z, &chaine,
																	&miroir);
		facette[count] = Facette( Point(x1,y1,z1),Point(x2,y2,z2),Point(x3,y3,z3), Coul , miroir );
	}


	fclose(scenefile);	

	/*************************************************************
	 * Calcul des variables
	 *************************************************************/
	ima = creerimage(NP, NL);

	F = NP/2.0/(tan((ANGLE/2.0)*PI/180.0));

	Ooeil = Point(xoeil,yoeil,zoeil);
	Dest  = Point(xdest,ydest,zdest);

	// cration de la matrice de passage
	MatPassage = MatricePassage(Ooeil, Dest);
	
	// diverses initialisations
	Lum.Init();

	// nb de rflexions successives
	nbreflexionsuccesives = 3;

	/*************************************************************
	 * PARCOURS de la grille d'cran
	 *
	 * Pour chaque ligne et chaque colone de l'cran
	 * a. Calcul du point destination Pij sur la grille
	 * b. Passer le point destination Pij dans le repre global
	 * c. Calculer la direction du vecteur Rd=(Oeil->Pij)  (normalises)
	 *                                          ^    ^
	 *                                          |    |
	 *                                    Coordones globales.
	 *
	 * Ecran de rsolution (NP, NL) = (500, 500) origine  (250,250)
	 * Ecran plac  distance F (focale)
	 *
	 * On place une sphre dans le repre global, et on cherche
	 * les intersections entre les rayons et la sphre (idem pour
	 * les autres formes).
	 *************************************************************/
	// Barre d'avancement du calcul
	printf( "Processing image. Please wait..."
			"\n"
			"\n0 [                                                  ] %d"
		    "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
		    "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", NP);


    for(int i=-(NL/2); i<(NL/2);i++)
	{
		// Avancement de la barre de progression
		if( (i+NL)%(NL/50) == 0 )
			printf("#");

		for(int j=-(NP/2); j<(NP/2);j++)
		{
			// rayon dans le repre de l'oeil (local)
			PtEcranLocal = Point(i,j,F);
			PtEcranGlobal = MatPassage.Mul(PtEcranLocal);

			Vtmp = Vecteur(Ooeil, PtEcranGlobal);
			Vtmp.Normaliser();
			Ray = Rayon(Ooeil, Vtmp);

			Lum = raytrace(Ray,nbreflexionsuccesives);

			ima.buffer[(i+NP/2)*3+(j+NL/2)*NP*3]   = (unsigned char)((int)(Lum.Ir));	//R
			ima.buffer[(i+NP/2)*3+(j+NL/2)*NP*3+1] = (unsigned char)((int)(Lum.Ig));	//G
			ima.buffer[(i+NP/2)*3+(j+NL/2)*NP*3+2] = (unsigned char)((int)(Lum.Ib));	//B
		}
		
	}

	res = ecrireimage(&ima, argimage);
	if( res == 1 )
		cout << "Succeded!" << "\n\n";
	else
		cout << "Failed!" << "\n\n";

	return(0);
}


ObjetIntersection CalculIntersections(Rayon Ray,Plan *P, Sphere *S, Facette *F)
{
	int i;
	ObjetIntersection I;

	I.zmin = 1e99;

	for( i=0;i<nbplans;i++)		P[i].Intersection(Ray, I);
	for( i=0;i<nbspheres;i++)	S[i].Intersection(Ray, I);
	for( i=0;i<nbfacettes;i++)	F[i].Intersection(Ray, I);

	I.Pinter = Point( Ray.porig.x + Ray.vdir.x * I.zmin,
					  Ray.porig.y + Ray.vdir.y * I.zmin,
					  Ray.porig.z + Ray.vdir.z * I.zmin );


	return(I);
}


ObjetIntersection CalculIntersectionsOmbres(Rayon Ray, Sphere *S, Facette *F)
{
	int i;
	ObjetIntersection I;

	I.zmin = 1e99;

	for( i=0;i<nbspheres;i++)	S[i].Intersection(Ray, I);
	for( i=0;i<nbfacettes;i++)	F[i].Intersection(Ray, I);
	
	I.Pinter = Point( Ray.porig.x + Ray.vdir.x * I.zmin,
					  Ray.porig.y + Ray.vdir.y * I.zmin,
					  Ray.porig.z + Ray.vdir.z * I.zmin );

	return(I);
}

Luminosite raytrace(Rayon Ray,int cont)
{
	if( (--cont) == -2 )
	{
		Luminosite Lum;

		Lum.Init();
		return(Lum);
	}

	Rayon RayOmbre,RayMiroir;
	ObjetIntersection I,I2,I3;
	Luminosite Lum,LumRes,LumBuffer,LumReflechie;
	Point source_penombre[7];
	Vecteur Vref;
	double delta,costheta;

	I = CalculIntersections(Ray,plan,sphere,facette);

	delta = 0.1;

	if(I.zmin >= 0 )
	{
		Lum.Init();
		LumBuffer.Init();
		LumReflechie.Init();

		for( int k=0; k<nbsources; k++ )
		{
			// pour la source courante, on cre 8 sources virtuelles...
			for (int l=0; l<8; l++)
				source_penombre[l] = source[k];

			/* ... que l'on dcale de delta
			 * suivant les directions  z constant
			 *
			 *				s1	s2	s3
			 *					
			 *	s0	====>	s8		s4
			 *				
			 *				s7	s6	s5
			 */
			// s1
			source_penombre[0].x += -delta;
			source_penombre[0].y += delta;
			// s2
			source_penombre[1].y += delta;
			// s3
			source_penombre[2].x += delta;
			source_penombre[2].y += delta;
			// s4
			source_penombre[3].x += delta;
			// s5
			source_penombre[4].x += delta;
			source_penombre[4].y += -delta;
			// s6
			source_penombre[5].y += -delta;
			// s7
			source_penombre[6].x += -delta;
			source_penombre[6].y += -delta;
			// s8
			source_penombre[7].x += -delta;

			// et c'est avec ces 8 sources que l'on calcule la pnombre
			for( int m=0; m<8; m++ )
			{
				RayOmbre.vdir = Vecteur(I.Pinter, source_penombre[m]);
				(RayOmbre.vdir).Normaliser();
				RayOmbre.porig = I.Pinter;
				I2 = CalculIntersectionsOmbres(RayOmbre,sphere,facette);

				if( I2.zmin == 1e99 )
					LumBuffer = LumBuffer.Add( Phong(Ray,I,source_penombre[m]) );
			}
			// on attnue les 8 sources
			LumBuffer.Ir /= 8;
			LumBuffer.Ig /= 8;
			LumBuffer.Ib /= 8;

			Lum = Lum.Add( LumBuffer );
		}
	}

	if( I.miroir == MIROIR )
	{
		// EFFET MIROIR
		// calcul du rayon rflchi par la surface
		costheta = (I.Vn).pScalaire(Ray.vdir);
		Vref.x = -(2.0*I.Vn.x*costheta - Ray.vdir.x);
		Vref.y = -(2.0*I.Vn.y*costheta - Ray.vdir.y);
		Vref.z = -(2.0*I.Vn.z*costheta - Ray.vdir.z);
		RayMiroir.vdir = Vref;
		(RayMiroir.vdir).Normaliser();
		RayMiroir.porig = I.Pinter;
		// appel rcursif  la fonction de lanc de rayon
		LumReflechie = raytrace(RayMiroir,cont);
	}

	// Somme de toutes les intensits lumineuse
	LumRes.Ir = I.Couleur.x * Lum.Ir + LumReflechie.Ir;	//R
	LumRes.Ig = I.Couleur.y * Lum.Ig + LumReflechie.Ig;	//G
	LumRes.Ib = I.Couleur.z * Lum.Ib + LumReflechie.Ib;	//B

	if( LumRes.Ir > 255.0)
		LumRes.Ir = 255.0;
	if( LumRes.Ir < 0.0)
		LumRes.Ir = 0.0;

	if( LumRes.Ig > 255.0)
		LumRes.Ig = 255.0;
	if( LumRes.Ig < 0.0)
		LumRes.Ig = 0.0;

	if( LumRes.Ib > 255.0)
		LumRes.Ib = 255.0;
	if( LumRes.Ib < 0.0)
		LumRes.Ib = 0.0;


	return(LumRes);
}