/*
 * Projet de lancer de rayons
 *
 * Jean-Guillaume Delculry
 *
 * STRUCTURES.CPP
 *
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "structures.h"


/**************************************
 * Fonctions de la classe POINT
 **************************************/
Point::Point()
{
	x = 0;
	y = 0;
	z = 0;
}

Point::Point(double nx, double ny, double nz)
{
	x = nx;
	y = ny;
	z = nz;
}


/**************************************
 * Fonctions de la classe VECTEUR
 **************************************/
Vecteur::Vecteur()  {
	x = 0;
	y = 0;
	z = 0;
}

Vecteur::Vecteur(double nx, double ny, double nz)  {
	x = nx;
	y = ny;
	z = nz;
}

Vecteur::Vecteur(Point a, Point b)  {
	x = b.x - a.x;
	y = b.y - a.y;
	z = b.z - a.z;
}

double Vecteur::pScalaire(Vecteur v)  {
	return( x * v.x +
			y * v.y +
			z * v.z);
}

Vecteur Vecteur::pVectoriel(Vecteur v)  {
        Vecteur vr;
        
		vr = Vecteur(y * v.z - z * v.y, z*v.x - x*v.z, x*v.y - y*v.x);

        return (vr);
}

double Vecteur::Norme()  {
	return( sqrt(x*x + y*y +z*z) );
}

void Vecteur::Normaliser()  {
	double norme;
	norme = Norme();

	if( norme != 0)  {
		x = x / norme;
		y = y / norme;
		z = z / norme;
	}
}


/**************************************
 * Fonctions de la classe RAYON
 **************************************/
Rayon::Rayon()  {
	porig = Point();
	vdir  = Vecteur();
}

Rayon::Rayon(Point a, Vecteur v)  {
	porig.x = a.x;
	porig.y = a.y;
	porig.z = a.z;

	vdir.x = v.x;
	vdir.y = v.y;
	vdir.z = v.z;
}


/**************************************
 * Fonctions de la classe MATRICE
 **************************************/
Matrice::Matrice()
{
	a = 0;
	b = 0;
	c = 0;
	d = 0;
	e = 0;
	f = 0;
	g = 0;
	h = 0;
	i = 0;
}

Matrice::Matrice(Vecteur U,Vecteur V,Vecteur W)
{
	a = U.x;
	b = V.x;
	c = W.x;
	d = U.y;
	e = V.y;
	f = W.y;
	g = U.z;
	h = V.z;
	i = W.z;
}

Matrice Matrice::Add(Matrice B)
{
	Matrice Buf;
	Buf = Matrice();

	Buf.a = a + B.a;
	Buf.b = b + B.b;
	Buf.c = c + B.c;
	Buf.d = d + B.d;
	Buf.e = e + B.e;
	Buf.f = f + B.f;
	Buf.g = g + B.g;
	Buf.h = h + B.h;
	Buf.i = i + B.i;

	return(Buf);
}

Point Matrice::Mul(Point p)
{
	// "Multiplication" d'une matrice 3x3 par un point 3x1
	// que l'on assimile alors  un vecteur
	Point Buf;
	Buf = Point();

	Buf.x = a*p.x + b*p.y + c*p.z;
	Buf.y = d*p.x + e*p.y + f*p.z;
	Buf.z = g*p.x + h*p.y + i*p.z;
	
	return(Buf);
}

/**************************************
 * Fonctions de la classe SPHERE
 **************************************/
Sphere::Sphere()
{
}

Sphere::Sphere(Point C,int m)
{
	porig.x = 0.0;
	porig.y = 0.0;
	porig.z = 0.0;

	Couleur = C;

	r = 0.0;

	miroir = m;
}

Sphere::Sphere(Point P, double rayon, Point C, int m)
{
	porig.x = P.x;
	porig.y = P.y;
	porig.z = P.z;

	Couleur = C;

	r = rayon;

	miroir = m;
}

void Sphere::Intersection(Rayon Ray, ObjetIntersection &I)
{
	double A,B,C,Det,Res1,Res2;
	double t;
	Vecteur Vnorm;

	A =	Ray.vdir.x*Ray.vdir.x +
		Ray.vdir.y*Ray.vdir.y +
		Ray.vdir.z*Ray.vdir.z;

	B = 2*(	Ray.vdir.x*(Ray.porig.x - porig.x) +
			Ray.vdir.y*(Ray.porig.y - porig.y) +
			Ray.vdir.z*(Ray.porig.z - porig.z));

	C =	(Ray.porig.x - porig.x)*(Ray.porig.x - porig.x) +
		(Ray.porig.y - porig.y)*(Ray.porig.y - porig.y) +
		(Ray.porig.z - porig.z)*(Ray.porig.z - porig.z) - r*r;

	Det = (B*B-4*A*C);

	if( Det>=0 )  
	{
		Res1 = (-B - sqrt(B*B-4*A*C))/2.0;
		Res2 = (-B + sqrt(B*B-4*A*C))/2.0;

		if( (Res1 > 0.0)&&(Res2 > 0.0) )
		{
			if( Res1 < Res2 )
				t = Res1;
			else
				t = Res2;

			if( (t>0.000001)&&((t<I.zmin)||(I.zmin==1e99)) )
			{
				// pour des raisons d'arrondis, il faut mettre 0.000001 au lieu de 0.0
				Vnorm = Vecteur( Ray.porig.x + Ray.vdir.x*t - porig.x,
				  				 Ray.porig.y + Ray.vdir.y*t - porig.y,
								 Ray.porig.z + Ray.vdir.z*t - porig.z );
				Vnorm.Normaliser();

				I.zmin = t;
				I.Vn = Vnorm;
				I.Couleur = Couleur;
				I.miroir = miroir;
			}
		}
	}
}


/**************************************
 * Fonctions de la classe PLAN
 **************************************/
Plan::Plan()
{
}

Plan::Plan(Point C, int m)
{
	vnorm.x = 0.0;
	vnorm.y = 0.0;
	vnorm.z = 0.0;

	Couleur = C;

	d = 0.0;

	miroir = m;
}


Plan::Plan(Vecteur V, double dist, Point C,int m)
{
	vnorm.x = V.x;
	vnorm.y = V.y;
	vnorm.z = V.z;

	Couleur = C;

	d = dist;

	miroir = m;
}

Plan::Plan(Point P, double dist, Point C,int m)
{
	vnorm.x = P.x;
	vnorm.y = P.y;
	vnorm.z = P.z;

	Couleur = C;

	d = dist;

	miroir = m;
}

void Plan::Intersection(Rayon Ray, ObjetIntersection &I)
{
	double vd,t;

	vd = vnorm.x*Ray.vdir.x +
		 vnorm.y*Ray.vdir.y +
		 vnorm.z*Ray.vdir.z;

	t = -(vnorm.x*Ray.porig.x +
		  vnorm.y*Ray.porig.y +
		  vnorm.z*Ray.porig.z + d)/vd;

	if( (t>0.00001)&&((t<I.zmin)||(I.zmin==1e99)) )
	{
		if( vd<0.0 )
		{
			I.zmin = t;
			I.Vn = vnorm;
			I.Couleur = Couleur;
			I.miroir = miroir;
		}
	}
}


/**************************************
 * Fonctions de la classe FACETTE
 **************************************/
Facette::Facette()
{
}

Facette::Facette(Point C,int m)
{
	Couleur = C;

	d = 0;

	miroir = m;
}

Facette::Facette(Point nA,Point nB,Point nC, Point Color,int m)
{
	Vecteur AB,AC;

	A = nA;

	B = nB;

	C = nC;

	Couleur = Color;

	miroir = m;

	// calcul du vecteur normal  la facette
	AB = Vecteur(nA,nB);
	AC = Vecteur(nA,nC);
	vnorm = AB.pVectoriel(AC);
	vnorm.Normaliser();

	d = -( vnorm.x*nA.x +
		   vnorm.y*nA.y +
		   vnorm.z*nA.z );
}

void Facette::Intersection(Rayon Ray, ObjetIntersection &I)
{
	
	double vd,t;
	Point Pint;
	Vecteur IA,IB,IC,vtmp1,vtmp2,vtmp3,vtmp4;
	double s1,s2;

	vd = vnorm.x*Ray.vdir.x +
		 vnorm.y*Ray.vdir.y +
		 vnorm.z*Ray.vdir.z;

	t = -(vnorm.x*Ray.porig.x +
		  vnorm.y*Ray.porig.y +
		  vnorm.z*Ray.porig.z + d)/vd;

	if( (t>0.00001)&&((t<I.zmin)||(I.zmin==1e99)) )
	{
		if( vd!=0.0 )
		{
			Pint = Point( Ray.porig.x + Ray.vdir.x*t,
						  Ray.porig.y + Ray.vdir.y*t,
						  Ray.porig.z + Ray.vdir.z*t);

			IA = Vecteur(Pint,A);
			IB = Vecteur(Pint,B);
			IC = Vecteur(Pint,C);

			vtmp1 = IA.pVectoriel(IB);
			vtmp2 = IA.pVectoriel(IC);
			s1 = vtmp1.pScalaire(vtmp2);

			if(s1 < 0)
			{
				vtmp3 = IB.pVectoriel(IA);
				vtmp4 = IB.pVectoriel(IC);

				s2 = vtmp3.pScalaire(vtmp4);

				if( (s2 < 0))
				{
					I.zmin = t;
					I.Vn = vnorm;
					I.Couleur = Couleur;
					I.miroir = miroir;
				}
			}
		}
	}
}

/**************************************
 * Fonctions de la classe INTERSECTION
 **************************************/
ObjetIntersection::ObjetIntersection()
{
}

void ObjetIntersection::Init()
{
	zmin = 0;
	Pinter = Point(0,0,0);
	Vn = Vecteur(0,0,0);
	Couleur = Point(0,0,0);
	miroir = 0;
}


/**************************************
 * Fonctions de la classe LUMINOSITE
 **************************************/
Luminosite::Luminosite()
{
}

void Luminosite::Init()
{
	Ir = 0;
	Ig = 0;
	Ib = 0;
}

Luminosite Luminosite::Add(Luminosite L)
{
	Luminosite Lres;

	Lres.Ir = Ir + L.Ir;
	Lres.Ig = Ig + L.Ig;
	Lres.Ib = Ib + L.Ib;

	return(Lres);
}


/**************************************
 * Autres fonctions
 **************************************/
Image creerimage(int NP, int NL)
{
	Image ima;
	ima.buffer = (unsigned char *)malloc(sizeof(unsigned char)*NP*NL*3);
	ima.np = NP;
	ima.nl = NL;
	return(ima);
}


int ecrireimage(Image *ima, char *name)  
{
	FILE *file ;
	int res;
	
	unlink(name);

	if(file= fopen(name,"wb"))
	{
		cout << "\n\nWriting RES.PPM to disk... ";
		fprintf(file, "P6\n%d %d\n255\n", ima->np, ima->nl);
		fwrite(ima->buffer, ima->np*ima->nl*3*sizeof(unsigned char),1,file);
		res = 1;
	}
	else 
	{
		cout << "I/O Problem" << endl;
		res = 0;
	}
	
	fclose(file);
	return(res);
}

Matrice MatricePassage(Point Origine_Oeil, Point Point_Regarde)
{
	Matrice MatRes;

	Vecteur U,V,W;

	V = Vecteur(0,1,0);
	
	W = Vecteur(Origine_Oeil, Point_Regarde);
	W.Normaliser();

	U = W.pVectoriel(V);
	U.Normaliser();

	V = W.pVectoriel(U);
	
	MatRes = Matrice(U,V,W);
	
	return( MatRes );
}


Luminosite Phong(Rayon Ray, ObjetIntersection &I, Point source)
{
	double ECLAIRAGE = 0.3;
	double N = 9; // rugosit
	double KS = 0.55;
	double KD = 0.45;
	double fa = 1.0;
	Point Ip = Point(1.0, 1.0, 1.0);

	double tmp,costheta = 0;
	Luminosite Lum;
	Vecteur Vr,Vn,Vl,Vv;

	// Calcul des vecteurs N, L et V
	Vn = I.Vn;
	Vl = Vecteur(I.Pinter, source);
	Vv = Vecteur(I.Pinter,Ray.porig);
	// Normalisation
	Vn.Normaliser();
	Vl.Normaliser();
	Vv.Normaliser();

	// Calcul de cos(thta)
	costheta = Vn.pScalaire(Vl);

	// Calcul de R
	Vr.x = 2.0*Vn.x*costheta - Vl.x;
	Vr.y = 2.0*Vn.y*costheta - Vl.y;
	Vr.z = 2.0*Vn.z*costheta - Vl.z;
	Vr.Normaliser();

	tmp = Vr.pScalaire(Vv);
	
	// Calcul de l'intensit I (sachant que IaKa = 0)
	Lum.Ir = fa * Ip.x * (KD*1.0*costheta + 1.0*KS*pow(tmp,N)) + ECLAIRAGE;
	Lum.Ig = fa * Ip.y * (KD*0.0*costheta + 1.0*KS*pow(tmp,N)) + ECLAIRAGE;
	Lum.Ib = fa * Ip.z * (KD*0.0*costheta + 1.0*KS*pow(tmp,N)) + ECLAIRAGE;

	return(Lum);
}
